@@ -3,7 +3,7 @@
    Lexical scanner for dhcpd config file... */
 
 /*
- * Copyright (c) 2004-2010 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2012 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -35,14 +35,14 @@
 #include "dhcpd.h"
 #include <ctype.h>
 
-static int get_char PROTO ((struct parse *));
+static int get_char (struct parse *);
 static void unget_char(struct parse *, int);
-static void skip_to_eol PROTO ((struct parse *));
+static void skip_to_eol (struct parse *);
 static enum dhcp_token read_whitespace(int c, struct parse *cfile);
-static enum dhcp_token read_string PROTO ((struct parse *));
-static enum dhcp_token read_number PROTO ((int, struct parse *));
-static enum dhcp_token read_num_or_name PROTO ((int, struct parse *));
-static enum dhcp_token intern PROTO ((char *, enum dhcp_token));
+static enum dhcp_token read_string (struct parse *);
+static enum dhcp_token read_number (int, struct parse *);
+static enum dhcp_token read_num_or_name (int, struct parse *);
+static enum dhcp_token intern (char *, enum dhcp_token);
 
 isc_result_t new_parse (cfile, file, inbuf, buflen, name, eolp)
 	struct parse **cfile;
@@ -999,6 +999,8 @@ intern(char *atom, enum dhcp_token dfv) {
 		if (!strncasecmp(atom + 1, "et", 2)) {
 			if (!strcasecmp(atom + 3, "-lease-hostnames"))
 				return GET_LEASE_HOSTNAMES;
+			if (!strcasecmp(atom + 3, "hostbyname"))
+				return GETHOSTBYNAME;
 			if (!strcasecmp(atom + 3, "hostname"))
 				return GETHOSTNAME;
 			break;
@@ -1041,6 +1043,8 @@ intern(char *atom, enum dhcp_token dfv) {
 			return INCLUDE;
 		if (!strcasecmp (atom + 1, "nteger"))
 			return INTEGER;
+		if (!strcasecmp (atom  + 1, "nfiniband"))
+			return TOKEN_INFINIBAND;
 		if (!strcasecmp (atom + 1, "nfinite"))
 			return INFINITE;
 		if (!strcasecmp (atom + 1, "nfo"))
@@ -1051,6 +1055,8 @@ intern(char *atom, enum dhcp_token dfv) {
 			return IP6_ADDRESS;
 		if (!strcasecmp (atom + 1, "nitial-interval"))
 			return INITIAL_INTERVAL;
+                if (!strcasecmp (atom + 1, "nitial-delay"))
+                        return INITIAL_DELAY;
 		if (!strcasecmp (atom + 1, "nterface"))
 			return INTERFACE;
 		if (!strcasecmp (atom + 1, "dentifier"))
@@ -1224,6 +1230,8 @@ intern(char *atom, enum dhcp_token dfv) {
 			return PEER;
 		if (!strcasecmp (atom + 1, "rimary"))
 			return PRIMARY;
+		if (!strcasecmp (atom + 1, "rimary6"))
+			return PRIMARY6;
 		if (!strncasecmp (atom + 1, "artner", 6)) {
 			if (!atom [7])
 				return PARTNER;
@@ -1313,7 +1321,9 @@ intern(char *atom, enum dhcp_token dfv) {
 			    tolower((unsigned char)atom[2]) == 'c') {
 				if (!strncasecmp(atom + 3, "ond", 3)) {
                                         if (!strcasecmp(atom + 6, "ary"))
-                                                return SECONDARY;
+						return SECONDARY;
+                                        if (!strcasecmp(atom + 6, "ary6"))
+						return SECONDARY6;
                                         if (!strcasecmp(atom + 6, "s"))
                                                 return SECONDS;
 					break;
